@@ -3,7 +3,7 @@
    DHCP Server Daemon. */
 
 /*
- * Copyright (c) 2004-2006 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2008 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1996-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -34,11 +34,11 @@
 
 #ifndef lint
 static char ocopyright[] =
-"$Id: dhcpd.c,v 1.115.2.19 2007/05/01 20:42:56 each Exp $ Copyright 2004-2006 Internet Systems Consortium.";
+"$Id: dhcpd.c,v 1.115.2.23 2008/05/14 20:54:24 dhankins Exp $ Copyright 2004-2008 Internet Systems Consortium.";
 #endif
 
   static char copyright[] =
-"Copyright 2004-2007 Internet Systems Consortium.";
+"Copyright 2004-2008 Internet Systems Consortium.";
 static char arr [] = "All rights reserved.";
 static char message [] = "Internet Systems Consortium DHCP Server";
 static char url [] = "For info, please visit http://www.isc.org/sw/dhcp/";
@@ -203,7 +203,6 @@ int main (argc, argv, envp)
 	int cftest = 0;
 	int lftest = 0;
 #ifndef DEBUG
-	int pidfilewritten = 0;
 	int pid;
 	char pbuf [20];
 	int daemon = 1;
@@ -538,28 +537,32 @@ int main (argc, argv, envp)
 
 	/* Read previous pid file. */
 	if ((i = open (path_dhcpd_pid, O_RDONLY)) >= 0) {
-		status = read (i, pbuf, (sizeof pbuf) - 1);
+		status = read(i, pbuf, (sizeof pbuf) - 1);
 		close (i);
 		if (status > 0) {
-			pbuf [status] = 0;
-			pid = atoi (pbuf);
+			pbuf[status] = 0;
+			pid = atoi(pbuf);
 
-			/* If the previous server process is not still running,
-			   write a new pid file immediately. */
-			if (pid && (pid == getpid() || kill (pid, 0) < 0)) {
-				unlink (path_dhcpd_pid);
-				if ((i = open (path_dhcpd_pid,
-					O_WRONLY | O_CREAT, 0644)) >= 0) {
-				    sprintf (pbuf, "%d\n", (int)getpid ());
-				    write (i, pbuf, strlen (pbuf));
-				    close (i);
-				    pidfilewritten = 1;
-				}
-			} else
-				log_fatal ("There's already a DHCP server running.");
+			/*
+                         * If there was a previous server process and it's
+                         * is still running, abort
+                         */
+			if (!pid || (pid != getpid() && kill(pid, 0) == 0))
+				log_fatal("There's already a "
+                                          "DHCP server running.");
 		}
 	}
 
+        /* Write new pid file. */
+        if ((i = open(path_dhcpd_pid, O_WRONLY|O_CREAT|O_TRUNC, 0644)) >= 0) {
+                sprintf(pbuf, "%d\n", (int) getpid());
+                write(i, pbuf, strlen(pbuf));
+                close(i);
+        } else {
+                log_error("Can't create PID file %s: %m.", path_dhcpd_pid);
+        }
+
+
 	/* If we were requested to log to stdout on the command line,
 	   keep doing so; otherwise, stop. */
 	if (log_perror == -1)
@@ -574,21 +577,6 @@ int main (argc, argv, envp)
 		close (2);
 		pid = setsid ();
 	}
-
-	/* If we didn't write the pid file earlier because we found a
-	   process running the logged pid, but we made it to here,
-	   meaning nothing is listening on the bootp port, then write
-	   the pid file out - what's in it now is bogus anyway. */
-	if (!pidfilewritten) {
-		unlink (path_dhcpd_pid);
-		if ((i = open (path_dhcpd_pid,
-			       O_WRONLY | O_CREAT, 0644)) >= 0) {
-			sprintf (pbuf, "%d\n", (int)getpid ());
-			write (i, pbuf, strlen (pbuf));
-			close (i);
-			pidfilewritten = 1;
-		}
-	}
 #endif /* !DEBUG */
 
 #if defined (DEBUG_MEMORY_LEAKAGE) || defined (DEBUG_MALLOC_POOL) || \
@@ -598,10 +586,6 @@ int main (argc, argv, envp)
 	dmalloc_outstanding = 0;
 #endif
 
-#if defined (DEBUG_RC_HISTORY_EXHAUSTIVELY)
-	dump_rc_history ();
-#endif
-
 	omapi_set_int_value ((omapi_object_t *)dhcp_control_object,
 			     (omapi_object_t *)0, "state", server_running);
 
